from netgen.csg import *
from netgen.geom2d import *
from ngsolve import *

import numpy as np


variant = 2


# DEFINE GEOMETRY and initial mesh
d = 3
mesh = Mesh (unit_cube.GenerateMesh(maxh=0.3))
dirichlet_BC = [1,2,3,4,5,6]

# SOLVER PARAMETERS
p = 3                   # polynomial order
nlevels = 2             # J (height of mesh hierarchy)
niter = 30;             # number of iteration of the solver

weight1 = 1.*(d+1)

# DEFINE PROBLEM

# define FE space
fes = H1(mesh, order=p, dirichlet=dirichlet_BC, low_order_space=True)
u,v = fes.TnT()

# the right hand side
f = LinearForm(fes)
f += 32 * (y*(1-y)+x*(1-x)+z*(1-z)) * v * dx

# the bilinear-form
a = BilinearForm(fes, symmetric=True)
a += grad(u)*grad(v)*dx # Poisson problem

# DEFINE AUXILIARY FUNCTIONS
def VertexBlocks(mesh, fes):
    blocks = []
    freedofs = fes.FreeDofs()
    for v in mesh.vertices:
        for dof in fes.GetDofNrs(v):
            if freedofs[dof]:
                blocks.append({dof})
    return blocks

def VertexDofs(mesh, fes):
    vertexdofs = BitArray(fes.ndof)
    vertexdofs[:] = False
    for v in mesh.vertices:
        for dof in fes.GetDofNrs(v):
            vertexdofs[dof] = True
    vertexdofs &= fes.FreeDofs()
    return vertexdofs


# DEFINE MG SOLVER/PRECONDITIONER

class MultiGrid(BaseMatrix):
    def __init__ (self, bfa, coarsedofs, w1=1.):
        super(MultiGrid, self).__init__()
        self.bfa = bfa
        self.mats = [bfa.mat]
        self.smoothers = [ () ]
        self.projlevel = [ Projector(mask=coarsedofs, range=True) ]
        self.w1 = weight1
        self.invcoarseproblem = bfa.mat.Inverse(coarsedofs)

    def Update(self, dofs, projdofs):
        self.mats.append (self.bfa.mat)
        self.smoothers.append(self.bfa.mat.CreateBlockSmoother( dofs ) )
        self.projlevel.append(Projector(mask=projdofs, range=True))


    def Height(self):
        return self.bfa.mat.height
    def Width(self):
        return self.bfa.mat.width

    def Mult(self, b, x):
        self.MGM(len(self.mats)-1, b, x)

    def MGM(self, level, b, rho):
        #
        rho[:] = 0
        self.projlevel[level].Project(b)
        if level > 0:
            prol = self.bfa.space.Prolongation()
            nc = self.mats[level-1].height
            rhoprev = b.CreateVector()
            tmp = b.CreateVector()
            # NO PRE-SMOOTHING
            #
            # coarse grid correction
            tmp.data = b
            prol.Restrict(level, tmp)
            self.MGM(level-1, tmp[0:nc] , rhoprev[0:nc])
            prol.Prolongate(level, rhoprev)
            #
            # post-smoothing weighted (block)Jacobi
            if variant == 1:
                rho.data = self.smoothers[level] * b
                rho.data *= (1./self.w1)  #weightening
            else:
                b.data *= (1./self.w1)
                rho.data = self.smoothers[level] * b
            #
            rho.data += rhoprev
        else:
            # p1 problem on the coarsest mesh
            rho.data = self.invcoarseproblem * b



# initialize solver
print ("level 0")
print ("ndof = ", fes.ndof)
a.Assemble()
f.Assemble()

pre = MultiGrid(a, coarsedofs=VertexDofs(mesh, fes), w1=weight1)

# RUN THE EXPERIMENT
resnorm = np.zeros(niter+1)

with TaskManager():
    for level in (range(1,nlevels+1)):
        print ("level", level)
        mesh.Refine()
        fes.Update()
        print ("ndof = ", fes.ndof)
        a.Assemble()
        f.Assemble()
        blocks = VertexBlocks(mesh, fes)
        pre.Update(blocks, VertexDofs(mesh, fes))
        #
    #
    xapprox = GridFunction(fes)
    xexact = GridFunction(fes)
    res = f.vec.CreateVector()
    rhoalg = GridFunction(fes)
    Arho = rhoalg.vec.CreateVector()
    xexact.vec.data = a.mat.Inverse(fes.FreeDofs(), inverse="sparsecholesky") * f.vec
    #
    proj_freedofs = Projector(mask=fes.FreeDofs(), range=True)
    # initial (zero) guess and the residual
    xapprox.vec[:] = 0
    res.data = (f.vec - a.mat * xapprox.vec)
    proj_freedofs.Project(res)
    resnorm[0] = InnerProduct(res, res)
    print("init. residual norm (squared) = ", InnerProduct(res, res))
    #
    # iterate
    for iter in range(niter):
        rhoalg.vec.data = pre * res
        Arho.data = a.mat * rhoalg.vec
        lmbda_nominator = InnerProduct(res, rhoalg.vec)
        lmbda = lmbda_nominator/InnerProduct(rhoalg.vec, Arho)
        print("lambda = ", lmbda)
        rhoalg.vec.data *= lmbda
        xapprox.vec.data += rhoalg.vec
        res.data = (f.vec - a.mat * xapprox.vec)
        proj_freedofs.Project(res)
        resnorm[iter+1] = InnerProduct(res, res)
        print("residual norm (squared) = ", resnorm[iter+1])
    #

resnorm = resnorm/resnorm[0]
